#* Imports
import codecs
import subprocess
import sys
import os
import re
import collections

#* Globals
sc_hookfn = None

#* SSH
HOST = None
class hostname:
    def __init__(self, host):
        self._old_host = sys.modules["pycook.elisp"].HOST
        sys.modules["pycook.elisp"].HOST = host

    def __enter__(self):
        pass

    def __exit__(self, *args, **kwargs):
        sys.modules["pycook.elisp"].HOST = self._old_host

#* Functional
def position(item, lst, default=None):
    if item in lst:
        return lst.index(item)
    else:
        return default

def set_difference(lst1, lst2):
    s = set(lst2)
    return [x for x in lst1 if x not in s]

def find_if(pred, lst):
    for item in lst:
        if pred(item):
            return item

def position_if(pred, lst):
    for (i, item) in enumerate(lst):
        if pred(item):
            return i

def flatten(seq):
    """Flatten a list of lists into a list."""
    return [item for sublist in seq for item in sublist]

def partition(n, seq):
    return [seq[i:i + n] for i in range(0, len(seq), n)]

def group_by(f, lst):
    res = collections.OrderedDict()
    for x in lst:
        fx = f(x)
        if fx in res:
            res[fx].append(x)
        else:
            res[fx] = [x]
    return res

def delete(element, lst):
    return [x for x in lst if x != element]

def delete_dups(lst):
    seen = set()
    seen_add = seen.add
    return [x for x in lst if not (x in seen or seen_add(x))]

#* Sys
def top_level():
    f = sys._getframe()
    while f.f_back:
        f = f.f_back
    return f

def crash():
    tf = top_level()
    f = sys._getframe().f_back
    tf.f_globals["lnames"] = f.f_locals.keys()
    for (k, v) in f.f_locals.items():
        tf.f_globals[k] = v
    raise RuntimeError("locals stored to globals")

#* OS
def user_login_name():
    import getpass
    return getpass.getuser()

def emacsclient_eval(expr):
    e = re.sub('"', "\\\"", expr)
    return lf('emacsclient -e "{e}"')

def emacs_cook_script(fname):
    d_lib = locate_dominating_file(__file__, "lib")
    if d_lib:
        return expand_file_name("../cook/" + fname, d_lib)
    else:
        # sys.path has '', handle loading this package from git
        d_etc = locate_dominating_file(__file__, "etc")
        if fname == "cook.el":
            return expand_file_name("../" + fname, d_etc)
        else:
            return expand_file_name(fname, d_etc)

def emacs_batch_eval(expr):
    e = re.sub('"', "\\\"", expr)
    script_el = emacs_cook_script("scripts.el")
    return lf('emacs -batch -l {script_el} --eval "{e}"')

def eeval(s):
    return shell_command_to_string(emacsclient_eval(s))

def beval(s, init_file=None):
    s = re.sub('"', "\\\"", s)
    if init_file:
        init = "-l "+ init_file
    else:
        init = ""
    return shell_command_to_string(lf('emacs -batch {init} --eval "(print {s})"'))

#* Files
def default_directory():
    return os.getcwd()

def locate_dominating_file(f, n):
    if file_directory_p(f):
        d = f
    else:
        d = file_name_directory(expand_file_name(f))
    while d != "/":
        nd = os.path.join(d, n)
        if file_exists_p(nd):
            return nd
        d = file_name_directory(d)

def make_directory(d):
    """Work around Python2/3 `os.makedirs' incompat."""
    if not os.path.exists(d):
        os.makedirs(d)

def expand_file_name(f, directory=None):
    if HOST:
        if ":" in f:
            return f
        else:
            return HOST + ":" + f
    if not directory:
        directory = os.getcwd()
    else:
        directory = os.path.expanduser(directory)
    if re.match("^~", f):
        return os.path.expanduser(f)
    elif re.match("\\.\\./", f):
        return os.path.realpath(os.path.join(directory, f))
    else:
        return os.path.join(directory, f)

def file_name_sans_extension(f):
    return os.path.splitext(f)[0]

def file_name_directory(f):
    return os.path.dirname(f)

def file_name_nondirectory(f):
    return os.path.basename(f)

def parse_fname(fname):
    if not isinstance(fname, str):
        return fname
    elif fname[0] == ".":
        return (None, os.path.realpath(expand_file_name(fname)))
    elif ":" in fname:
        return fname.split(":")
    elif HOST is not None:
        return (HOST, fname)
    else:
        return (None, os.path.realpath(expand_file_name(fname)))

def file_exists_p(f):
    (host, fname) = parse_fname(f)
    if host is not None:
        with hostname(host):
            res = sc(
                "stat {fname} 2>/dev/null || echo Fail",
                desc=(host, "stat " + fname))
            return res != "Fail"
    else:
        return os.path.exists(expand_file_name(fname))

def file_newer_than_file_p(f1, f2):
    return os.path.getctime(f1) > os.path.getctime(f2)

def file_directory_p(f):
    return os.path.isdir(f)

def abbreviate_file_name(f, d):
    if not d[-1] == "/":
        d = d + "/"
    m = re.match(d, f)
    if m:
        return f[m.end():]
    else:
        m = re.match(f, d)
        if m:
            return ".".join(["../"]* (d[m.end():].count("/") - 1))

def directory_files(d, full=False, match=False):
    fs = os.listdir(d)
    if match:
        fs = [f for f in fs if re.search(match, f) is not None]
    if full:
        fs = [expand_file_name(f, d) for f in fs]
    return fs

def delete_file(f):
    return os.remove(f)

#* File read/write
def slurp(f):
    (host, fname) = parse_fname(f)
    if host:
        with hostname(host):
            return sc("cat {fname}")
    else:
        try:
            with open(expand_file_name(f), 'r') as fh:
                return fh.read()
        except UnicodeDecodeError:
            with codecs.open(
                    expand_file_name(f), 'r',
                    encoding="utf-8",
                    errors="ignore") as fh:
                return fh.read()

def slurp_lines(f):
    return slurp(f).splitlines()

def barf(f, s):
    f = os.path.expanduser(f)
    with open(f, 'w') as fh:
        fh.write(s)

#* Shell
def shell_command_to_string(cmd, **kwargs):
    if HOST:
        cmds = ["ssh", HOST, cmd]
    else:
        cmds = ["bash", "-c", cmd]
    out = subprocess.check_output(cmds, **kwargs).strip()
    if isinstance(out, str):
        return out
    else:
        return out.decode()

def sc(cmd, **kwargs):
    fcmd = lf(cmd, 2)
    if "desc" in kwargs:
        desc = kwargs["desc"]
        del kwargs["desc"]
    else:
        desc = None
    if sc_hookfn:
        sc_hookfn(fcmd, desc=desc)
    return shell_command_to_string(fcmd, **kwargs)

def shell_command_to_list(cmd, **kwargs):
    cmd_output = shell_command_to_string(cmd, **kwargs)
    return [s for s in cmd_output.split("\n") if s]

def sc_l(cmd, **kwargs):
    fcmd = lf(cmd, 2)
    if sc_hookfn:
        sc_hookfn(fcmd)
    return shell_command_to_list(fcmd, **kwargs)

def scb(cmd):
    return bash(lf(cmd, 2), capture=True)

def bash(cmd, echo=False, capture=False, **kwargs):
    if isinstance(cmd, list):
        cmd = "\n".join(cmd)
    if echo:
        sep = "-"*80
        print(sep)
        print("Run: \n" + cmd)
        print(sep)
    sys.stdout.flush()

    if HOST:
        cmds = ["ssh", HOST, cmd]
    else:
        cmds = ["/bin/bash", "-e", "-c", cmd]

    if "desc" in kwargs:
        desc = kwargs["desc"]
        del kwargs["desc"]
    else:
        desc = (HOST, cmd)

    if sc_hookfn:
        sc_hookfn(cmd, desc=desc)

    if capture:
        p = subprocess.Popen(cmds, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs)
        out = ""
        while True:
            part = p.stdout.read().decode()
            if part == "" and p.poll() is not None:
                break
            out += part
            if echo:
                print(part, end="")
        err = ""
        while True:
            part = p.stderr.read().decode()
            if part == "" and p.poll() is not None:
                break
            err += part
            if echo:
                print(part, end="")
        if p.returncode == 0:
            return err + out
        else:
            raise subprocess.CalledProcessError(p.returncode, cmd)
    else:
        p = subprocess.Popen(cmds, **kwargs)
        return_code = p.wait()
        if return_code == 0:
            return 0
        else:
            sys.stdout.flush()
            sys.stderr.flush()
            raise subprocess.CalledProcessError(return_code, cmd)

#* String
def lf(string, lvl=1):
    fr = sys._getframe()
    for _ in range(lvl):
        fr = fr.f_back
    vars_dict = fr.f_globals.copy()
    vars_dict.update(fr.f_locals)
    return string.format(**vars_dict)

#* Regex
def re_filter(regex, seq):
    return list(filter(lambda s: re.search(regex, s), seq))

def re_seq(regex, s):
    return re.findall(regex, s)

def replace_regexp_in_string(regexp, rep, string):
    return re.sub(re.compile(regexp, re.MULTILINE), rep, string)

#* Time
def timestamp():
    from datetime import datetime
    t = datetime.now()
    year = t.year
    month = t.month
    day = t.day
    hour = t.hour
    minute = t.minute
    dow = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"][t.weekday()]
    return lf("<{year}-{month:02d}-{day:02d} {dow} {hour:02d}:{minute:02d}>")



import pycook.elisp as el

def ls(recipe, package):
    if isinstance(recipe, int):
        return ["dpkg -L " + package]
    elif recipe[0] == "complete":
        part = recipe[1]
        if len(part) > 0:
            part += "*"
        return el.shell_command_to_string("dpkg-query -W -f '${Package}\n' " + part)


def file_to_package(recipe, fname):
    return ["dpkg -S " + fname]

def exe_to_package(recipe, exe):
    if type(recipe) is int:
        return el.lf("dpkg -S $(which {exe})")
    elif recipe[0] == "complete":
        return el.sc("compgen -c -A file " + recipe[1])



#* Imports
import re
import os
import shlex
import pycook.elisp as el
import pycook.recipes.git as git
from pycook.elisp import sc, lf, bash, parse_fname, scb, hostname
os.environ["TERM"] = "linux"

#* Functions
def package_installed_p_dpkg(package):
    res = sc("dpkg --get-selections '{package}'")
    if res == "" or re.search("deinstall$", res):
        return False
    else:
        return True

def package_installed_p_yum(package):
    res = scb("yum list installed '{package}' || true")
    return "Error: No matching Packages to list" not in res

def install_package(package, url=None):
    if isinstance(package, list):
        res = False
        for p in package:
            res |= install_package(p, url)
        return res
    user = sc("whoami")
    su = "" if user == "root" else "sudo "
    if el.file_exists_p("/usr/bin/dpkg"):
        if package_installed_p_dpkg(package):
            print(lf("{package}: OK"))
            return False
        else:
            if url is None:
                bash(lf("{su}apt-get update && DEBIAN_FRONTEND=noninteractive {su}apt-get install -y {package}"))
            else:
                fname = wget(url)
                bash(lf("{su}dpkg -i {fname}"))
            return True
    else:
        if package_installed_p_yum(package):
            print(lf("{package}: OK"))
            return False
        else:
            if url is None:
                bash(lf("{su}yum update -y && {su}yum upgrade -y && {su}yum install -y '{package}'"))
            else:
                fname = wget(url)
                bash(lf("{su}yum localinstall -y {fname}"))
            return True

def apt_key_add(email, url):
    if grep("apt-key list", email):
        print(email + " key: OK")
    else:
        bash(lf("wget -qO - {url} | apt-key add -"))

def debconf_select(package, var, v_type, v_val):
    bash(lf("echo '{package} {var} {v_type} {v_val}' | debconf-set-selections"))

def wget(url, download_dir="/tmp/"):
    if download_dir[:-1] == "/":
        fname = url.split("/")[-1]
        full_name = el.expand_file_name(fname, download_dir)
    else:
        full_name = download_dir
    if el.file_exists_p(full_name):
        print(lf("{full_name}: OK"))
    else:
        bash(lf("wget '{url}' -O {full_name}"))
        return full_name

def systemctl_start(service):
    if not el.scb("systemctl is-active {service} || true") == "active\n":
        el.bash(lf("systemctl start {service}"))
        return True
    else:
        print(lf("{service}: OK"))
        return False

def systemctl_enabled_services():
    cmd = "systemctl list-unit-files | grep enabled | grep service"
    lines = [re.split("[ \t]+", l)[0] for l in el.sc_l(cmd)]
    return [re.split("[.@]", l)[0] for l in lines]

def systemctl_enable(service):
    if service in systemctl_enabled_services():
        print(lf("{service}: OK"))
        return False
    else:
        el.bash(lf("systemctl enable {service}"))
        return True

def git_clone(url, target_dir, commit=None):
    target_dir = el.expand_file_name(target_dir)
    if el.file_exists_p(target_dir):
        print(lf("{target_dir}: OK"))
    else:
        gdir = el.expand_file_name(target_dir)
        pdir = el.file_name_directory(gdir)
        if not el.file_exists_p(pdir):
            el.make_directory(pdir)
        (_, gdir) = el.parse_fname(gdir)
        sc("git clone --recursive {url} {gdir}")
        if commit:
            sc("cd {gdir} && git reset --hard {commit}")

def symlink_p(fname):
    return " -> " in el.sc("stat {fname}")

def sudo(cmd, fname=None):
    if fname:
        if os.access(fname, os.W_OK):
            return cmd
        elif os.access(os.path.dirname(fname), os.W_OK | os.X_OK):
            return cmd
        else:
            return "sudo " + cmd
    user = sc("whoami")
    su = "" if user == "root" else "sudo "
    return su + cmd

def ln(fr, to):
    fr = el.expand_file_name(fr)
    if not el.file_exists_p(fr):
        raise RuntimeError("File doesn't exist", fr)
    to = el.expand_file_name(to)
    if el.file_directory_p(to) and not el.file_directory_p(fr):
        to_full = el.expand_file_name(el.file_name_nondirectory(fr), to)
    else:
        to_full = to
    if el.file_exists_p(to_full):
        if symlink_p(to_full):
            print(lf("{to_full}: OK"))
        else:
            if file_equal(fr, to_full):
                print(lf("{to_full} exists, contents equal"))
            else:
                print(lf("{to_full} exists, contents NOT equal"))
    else:
        fr_abbr = os.path.relpath(fr, os.path.dirname(to))
        cmd = sudo(lf("ln -s {fr_abbr} {to_full}"), to_full)
        bash(cmd)

def file_equal(f1, f2):
    def md5sum(f):
        return el.sc("md5sum " + shlex.quote(f)).split(" ")[0]
    return md5sum(f1) == md5sum(f2)

def cp_host(fr, to):
    # if el.file_exists_p(to) and file_equal(fr, to):
    #     print(lf("{to}: OK"))
    #     return False
    host = el.HOST
    with el.hostname(None):
        fr = el.expand_file_name(fr)
        el.sc("scp -r '{fr}' '{host}:{to}'")
        return True

def echo(fr_text, to):
    if el.file_exists_p(to) and fr_text == el.slurp(to):
        print(lf("{to}: OK"))
        return False
    elif el.HOST is None:
        el.sc(f"echo '{fr_text}' | " + sudo(f"tee {to}", to))
        return True
    else:
        host = el.HOST
        with el.hostname(None):
            el.sc(
                "echo '{fr_text}' | ssh '{host}' -T 'cat > {to}'",
                desc=(host, "write " + to))
            return True

def cp(fr, to):
    fr = el.expand_file_name(fr)
    to = el.expand_file_name(to)
    if el.file_exists_p(to) and file_equal(fr, to):
        print(lf("{to}: OK"))
        return False
    else:
        el.sc(sudo("cp '{fr}' '{to}'", to))
        return True

def chmod(fname, permissions):
    current = sc("stat -c %a {fname}")
    if current == permissions:
        print(lf("{fname}: OK"))
        return False
    else:
        if el.HOST is not None:
            scb("chmod {permissions} {fname}")
        else:
            cmd = sudo(lf("chmod {permissions} {fname}"), fname)
            bash(cmd)
        return True

def chown(fname, owner):
    current = sc("stat -c %U:%G {fname}")
    if current == owner:
        print(lf("{fname}: OK"))
        return False
    else:
        if el.HOST is not None:
            bash(lf("chown -R {owner} {fname}"))
        else:
            bash(lf("sudo chown -R {owner} {fname}"))
        return True

def barf(fname, text):
    if el.file_exists_p(fname) and text == el.slurp(fname):
        print(lf("{fname}: OK"))
    else:
        quoted_text = shlex.quote(text)
        bash(lf("echo {quoted_text} | sudo tee {fname} > /dev/null"))

def get_change_time(fname):
    fname = el.expand_file_name(fname)
    if el.file_directory_p(fname):
        git_dir = el.expand_file_name(".git", fname)
        if el.file_exists_p(git_dir):
            return float(git.mtime(fname).strftime("%s"))
        else:
            raise RuntimeError("Directory is not a git repo")
    else:
        return os.path.getctime(fname)

def make(target, cmds, deps=()):
    if (el.file_exists_p(target) and \
        all([get_change_time(target) > get_change_time(dep) for dep in deps])):
        print(lf("{target}: OK"))
        return False
    else:
        if isinstance(cmds, str):
            cmds = [cmds]
        elif callable(cmds):
            cmds(target)
            return True
        fcmds = []
        for cmd in cmds:
            cmd1 = re.sub("\\$@", target, cmd)
            cmd2 = re.sub("\\$\\^", " ".join([shlex.quote(dep) for dep in deps]), cmd1)
            cmd3 = re.sub("\\$<", shlex.quote(deps[0]), cmd2) if deps else cmd2
            if el.sc_hookfn:
                el.sc_hookfn(cmd3)
            fcmds.append(cmd3)
        bash(fcmds)
        return True

def curl(link, directory="~/Software"):
    fname = el.expand_file_name(link.split("/")[-1], directory)
    make(fname, "curl " + link + " -o " + shlex.quote(fname))
    return fname

def grep(cmd, regex):
    fcmd = cmd + " 2>/dev/null | grep -q '" + regex + "' && echo Y || echo N"
    return el.sc(fcmd) == "Y"

def render_patch(patch_lines, before):
    regex = "^\\+" if before else "^\\-"
    return "\n".join([line[1:] for line in patch_lines if not re.match(regex, line)])

def parse_patches(patches):
    if isinstance(patches, list):
        return [p if isinstance(p, str) else "\n".join(p) for p in patches]
    ls = el.slurp_lines(patches)
    res = []
    cur = []
    i = 0
    n = len(ls)
    while i < n:
        l = ls[i]
        if re.match("---|\+\+\+|@@", l):
            cur = []
            i += 1
        elif re.match(" ", l):
            cur.append(l)
            i += 1
        elif re.match("[+-]", l):
            cur.append(l)
            i += 1
            while re.match("[+-]", ls[i]) and i < n:
                cur.append(ls[i])
                i += 1
            res.append("\n".join(cur))
            cur = []
        elif l == "":
            i += 1
        else:
            raise RuntimeError("Unexpected", l)
    return res

def patch(fname, patches):
    """Patch FNAME applying PATCHES.
    Each PATCH in PATCHES is in diff -u format.
    Each line in PATCH starts with [+- ].
    [ ] lines are the optional start and end context.
    [-] lines are expected to exist in the file and will be removed.
    [+] lines will be added to the file after [-] lines are removed.

    PATCH lines should be in contiguous order: optional start context,
    [-] lines, [+] lines, optional end context.

    If PATCH was already applied for FNAME, it will be ignored.
    PATCHES can also be a file name of a "diff -u" output.
    """
    (host, name) = parse_fname(fname)
    patches = parse_patches(patches)
    if el.file_exists_p(fname):
        txt = el.slurp(fname)
    else:
        assert not any([
            re.search("^\\-", patch, flags=re.MULTILINE)
            for patch in patches])
        el.sc("touch {name}")
        txt = ""
    no_change = True
    for ptch in patches:
        patch_lines = el.delete("", ptch.splitlines())
        chunk_before = render_patch(patch_lines, True)
        chunk_after = render_patch(patch_lines, False)
        if chunk_before == "":
            if not re.search("^" + re.escape(chunk_after), txt, re.MULTILINE):
                no_change = False
                if not (txt == "" or txt[-1] == "\n"):
                    txt += "\n"
                txt += chunk_after + "\n"
        elif (re.search("^" + re.escape(chunk_before), txt, re.MULTILINE) and
              not re.search("^" + re.escape(chunk_after), txt, re.MULTILINE)):
            no_change = False
            txt = re.sub("^" + re.escape(chunk_before), chunk_after, txt, flags=re.MULTILINE)
        else:
            # either already patched or fail
            if not chunk_after in txt:
                raise RuntimeError("Patch failed", ptch)
    if no_change:
        print(fname + ": OK")
        return False
    else:
        print(fname + ": Patch")
        with el.hostname(None):
            el.barf("/tmp/insta.txt", txt)
        if host is not None:
            cmd = lf("scp /tmp/insta.txt {host}:{name}")
        else:
            cmd = sudo(lf("cp /tmp/insta.txt {name}"), name)
        with el.hostname(None):
            bash(cmd, desc=(host, "edit " + name))
        return True



#* Includes
import sys
import os
import re
import subprocess
import imp
import ast
import inspect
import collections
import pycook.elisp as el
import pycook.recipes as recipes
lf = el.lf

#* Globals
start_dir = el.default_directory()

#* Functions
def recipe_p(x):
    try:
        return inspect.getfullargspec(x[1]).args[0] == "recipe"
    except:
        return None

def recipe_names_ordered(book):
    body = ast.parse(el.slurp(book)).body
    return [f.name for f in body if isinstance(f, ast.FunctionDef)]

def recipe_dict(book):
    mod = imp.load_source("Cookbook", book)
    funs = inspect.getmembers(mod, inspect.isfunction)
    funs = filter(recipe_p, funs)
    names = recipe_names_ordered(book)
    items = sorted(funs, key=lambda x: el.position(x[0], names, 42))
    return collections.OrderedDict(items)

def book_config(book):
    rc_file = el.expand_file_name("~/.cook.d/__config__.py")
    if el.file_exists_p(rc_file):
        mod = imp.load_source("book_config", rc_file)
        config = mod.config
        if book in config:
            return config[book]
        elif "*" in config:
            return config["*"]
    return {}

def recipe_args_description(f):
    spec = inspect.getfullargspec(f)
    res = []
    ld = len(spec.defaults) if spec.defaults else 0
    d = len(spec.args) - ld - 1

    for (i, a) in enumerate(spec.args[1:]):
        if i >= d:
            res.append(":" + a + "=" + str(spec.defaults[i - d]))
        else:
            res.append(":" + a)
    return " " + " ".join(res)

def recipe_names(book):
    di = recipe_dict(book)
    ns = [k + recipe_args_description(v) for (k, v) in di.items()]
    return "\n".join(ns)

def describe(book, module=""):
    res = "Usage: cook [options] [book]"
    if module:
        res += " :" + module
    res += " <recipe>\n"
    res += "\nBook: " + book + "\n"
    names = "\n".join(["  " + n for n in recipe_names(book).split("\n")])
    return res + "\nRecipes:\n" + names + """

Options:
  -h, --help                  Show help.
  --list                      List only the recipes, separated by newlines.
  -p                          Print commands instead of running them."""

def script_get_book():
    dd = el.default_directory()
    d1 = el.locate_dominating_file(dd, "Cookbook.py")
    d2 = el.locate_dominating_file(dd, "cook/Cookbook.py")
    if d1 and (not d2 or len(d1) > len(d2) - 5):
        return (d1, el.file_name_directory(d1))
    elif d2:
        return (d2, el.file_name_directory(el.file_name_directory(d2)))
    else:
        raise RuntimeError("No Cookbook.py or cook/Cookbook.py found")

def get_book():
    caller_frame = sys._getframe().f_back.f_back
    caller_file = caller_frame.f_code.co_filename
    book = os.path.realpath(caller_file)
    return book

def log_file_name(base_dir, book, recipe):
    sub_dir = "_".join(el.delete("", os.path.normpath(book).split(os.sep)[:-1]))
    full_dir = el.expand_file_name(sub_dir, base_dir)
    el.make_directory(full_dir)
    ts = el.replace_regexp_in_string(" ", "_", el.timestamp())
    name = el.lf("{ts}-{recipe}.txt")
    return el.expand_file_name(name, full_dir)

def function_arglist(f):
    spec = inspect.getfullargspec(f)
    if spec.varargs:
        return [*spec.args, spec.varargs]
    else:
        return spec.args

def recipe_arity(f):
    spec = inspect.getfullargspec(f)
    args = [*spec.args, spec.varargs] if spec.varargs else spec.args
    res = len(args) - 1
    if spec.defaults:
        res -= len(spec.defaults)
    return res

class CommandLog:
    def __init__(self):
        self.cmds = []
        self.current_host = None

    def record(self, cmd, desc=None):
        if desc:
            (host, cmd) = desc
        else:
            host = el.HOST
        m = re.match("^cat (.*)$", cmd)
        if m:
            if self.cmds and self.cmds[-1] == "# stat " + m.group(1):
                self.cmds.pop()
        if host != self.current_host:
            self.current_host = host
            if host:
                self.cmds.append("# ssh " + host)
            else:
                self.cmds.append("# exit")
        self.cmds.append("# " + re.sub("\n", "\\\\n", cmd))

def _main(book, module, flags, args):
    if len(args) == 0:
        print(describe(book, module))
    elif args[0] == "--list":
        print(recipe_names(book))
    elif args[0] == "--help":
        print(describe(book))
    else:
        recipe = args[0]
        fun = recipe_dict(book)[recipe]
        cfg = book_config(book)
        if "tee" in cfg and recipe != "bash":
            basedir = os.path.expanduser(cfg["tee"]["location"])
            fname = log_file_name(basedir, book, recipe)
            el.barf(fname, lf("Book: {book}\nRecipe: {recipe}\n"))
            tee = subprocess.Popen(["setsid", "-w", "tee", "-a", fname], stdin=subprocess.PIPE)
            os.dup2(tee.stdin.fileno(), sys.stdout.fileno())
            os.dup2(tee.stdin.fileno(), sys.stderr.fileno())

        old_sc_hookfn = el.sc_hookfn
        log = CommandLog()
        el.sc_hookfn = log.record
        try:
            if "-p" in flags:
                sys.stdout = open(os.devnull, "w")
            ret_cmds = fun(42, *recipe_args(fun, args[1:])) or []
        except:
            if cfg.get("pdb", False):
                import pdb
                pdb.set_trace()
                return
            else:
                raise
        el.sc_hookfn = old_sc_hookfn
        if "-l" in flags:
            sys.stdout = sys.__stdout__
            print("\n".join([cmd[2:] for cmd in log.cmds] + ret_cmds))
        else:
            if len(log.cmds) < 10:
                spec = inspect.getfullargspec(fun)
                if ("log" in spec.args) and spec.defaults and spec.defaults[spec.args.index("log") - 1] is None:
                    pass
                else:
                    print("\n".join(log.cmds))
            if ret_cmds:
                el.bash(ret_cmds, echo=True)

def modules(full=False, match=False):
    cook_dir = el.file_name_directory(recipes.__file__)
    cook_modules = el.directory_files(cook_dir, full, match)
    user_dir = el.expand_file_name("~/.cook.d")
    if el.file_exists_p(user_dir):
        df = el.directory_files(user_dir, full, match)
        user_modules = [
            f for f in df
            if os.path.isfile(el.expand_file_name(f, user_dir))
            and os.path.splitext(f)[1] == ".py"]
    else:
        user_modules = []
    cook_modules += user_modules
    return list(filter(lambda s: not re.search("__", s), cook_modules))

def module_names():
    ms = modules(False, "[^_]\\.py$")
    return el.delete_dups([s[:-3] for s in ms])

def recipe_args(f, args_provided):
    spec = inspect.getfullargspec(f)
    args_req = spec.args
    assert args_req[0] == "recipe"
    args_missing = args_req[1 + len(args_provided):]
    if spec.defaults:
        args_missing = args_missing[:-len(spec.defaults)]
    args_input = []
    for a in args_missing:
        args_input.append(input(a + ": "))
    return args_provided + args_input

def parse_flags(argv):
    res = []
    i = 1
    for x in argv[1:]:
        if re.match("-[^-]", x):
            res.append(x)
            i += 1
        else:
            break
    return (res, argv[i:])

def main(argv=None):
    if argv is None:
        argv = sys.argv
    try:
        (flags, rest) = parse_flags(argv)
        if rest == [":"]:
            for module in modules(True):
                print(module)
            sys.exit(0)
        if len(rest) >= 1 and re.match("^:", rest[0]):
            module = rest[0][1:]
            book = get_module(module)
            args = rest[1:]
        else:
            module = ""
            (book, dd) = script_get_book()
            os.chdir(dd)
            args = rest
        _main(book, module, flags, args)
    except subprocess.CalledProcessError as e:
        # print(e)
        sys.exit(e.returncode)
    except KeyboardInterrupt:
        sys.exit(1)
    except RuntimeError as e:
        print(e)
        sys.exit(1)

def get_module(name):
    """Load a module NAME.
    If two modules are on sys.path, prefer the one on ~/.cook.d/.
    """
    mods = modules(True, name)
    if len(mods) == 2:
        return el.re_filter("\\.cook\\.d/", mods)[0]
    else:
        assert len(mods) == 1, mods
        return mods[0]

def complete(argv=None):
    if argv is None:
        argv = sys.argv
    assert argv[1] == "cook"
    (_flags, args) = parse_flags(argv[1:])

    # below, assume we're completing the last word
    # current word being completed is sys.argv[-1]
    args = args[:-1]

    # fix the difference between bash-completion.el and the actual bash completion
    if re.match(":.", args[0]):
        args = [":", args[0][1:]] + args[1:]

    if len(args) == 1:
        if args[0] == ":":
            print("\n".join(module_names()))
        else:
            rs = el.sc("cook --list").split("\n")
            # remove extra args
            rs = [re.split(" :", s)[0] for s in rs]
            fr = [r for r in rs if re.match(args[0], r)]
            print("\n".join(fr))
    elif len(args) == 2 and args[0] == ":":
        matching_cands = el.re_filter("^" + args[1], module_names())
        print("\n".join(matching_cands))
    elif len(args) == 3 and args[0] == ":":
        mod = get_module(args[1])
        cands = list(recipe_dict(mod).keys())
        matching_cands = el.re_filter("^" + args[2], cands)
        print("\n".join(matching_cands))
    elif len(args) >= 4 and args[0] == ":":
        mod = get_module(args[1])
        fun = recipe_dict(mod)[args[2]]
        part = args[-1]
        fun_args = function_arglist(fun)
        arg_idx = len(args) - 3
        if arg_idx < len(fun_args) and fun_args[arg_idx] in ["fname", "fnames"]:
            print(el.sc("compgen -f -- {part}"))
        else:
            args = [""]*recipe_arity(fun)
            print(fun(("complete", part), *args))
    else:
        print("")



def http_server(recipe):
    return [
        "python -m SimpleHTTPServer &",
        "python_process=$!",
        "sleep 0.5 && xdg-open http://localhost:8000",
        "wait $python_process",
    ]



#* Imports
import pycook.insta as st

#* Recipes
def enddate(recipe, fname):
    return "openssl x509 -enddate -noout -in " + fname

def txt(recipe, fname):
    return "openssl x509 -text -in " + fname

def pfx_to_pem(recipe, fname):
    """FNAME is a pfx file."""
    f_key = "privatekey.pem"
    f_cert = "publiccert.pem"
    f_pem = "intermediate.pem"
    f_ppem = "final.pem"
    st.make(f_key, f"openssl pkcs12 -in {fname} -nocerts -out {f_key} -nodes")
    st.make(f_cert, f"openssl pkcs12 -in {fname} -nokeys -out {f_cert} -nodes")
    st.make(f_pem, f"cat {f_key} {f_cert} > {f_pem}")
    st.make(f_ppem, [
        f"openssl rsa -in {f_pem} -out {f_ppem}",
        f"openssl x509 -in {f_pem} >> {f_ppem}"
    ])



def ls_users(recipe):
    return "cut -d: -f1 /etc/group | sort"

def ls_groups(recipe):
    return "cut -d: -f1 /etc/group | sort"

def ls_cpus(recipe):
    return "cat /proc/cpuinfo | awk '/^processor/{print $3}'"
